﻿using System;
using System.Collections.Generic;
using System.Text;
using System.ServiceModel;
using System.Timers;
using System.ServiceModel.Description;

namespace ServerTimeServiceApp
{
	[ServiceContract]
	interface IServerTimeNotification
	{
		[OperationContract(IsOneWay = true)]
		void OnTick(DateTime serverTime);
	}

	[ServiceContract(CallbackContract = typeof(IServerTimeNotification), 
					SessionMode = SessionMode.Required)]
	interface IServerTime
	{
		[OperationContract]
		void Subscribe();

		[OperationContract]
		void Unsubscribe();
	}

	[ServiceBehavior(InstanceContextMode = InstanceContextMode.PerSession,
		ConcurrencyMode = ConcurrencyMode.Single)]
	class ServerTimeService : IServerTime
	{
		private static Dictionary<string, IServerTimeNotification> _Clients;
		private static Timer _Timer;

		public static void StartService()
		{
			_Clients = new Dictionary<string,IServerTimeNotification>();
			_Timer = new Timer();

			_Timer.Interval = 1000;
			_Timer.Elapsed += delegate(object sender, ElapsedEventArgs e)
			{
				List<string> errorList = new List<string>();
				DateTime now = DateTime.Now;
				// 이벤트 발생도중 리스트에 추가/삭제를 막기 위해 잠금 사용
				lock (_Clients) {
					foreach (IServerTimeNotification callback in _Clients.Values) {
						IClientChannel channel = callback as IClientChannel;
						if (channel.State == CommunicationState.Opened) {
							callback.OnTick(now);
						}
						else {
							Console.WriteLine("Session Closed : {0}", channel.SessionId);
							errorList.Add(channel.SessionId);
						}
					}
					foreach (string sessionId in errorList) {
						_Clients.Remove(sessionId);
					}
				}
			};
			_Timer.Enabled = true;
		}

		public static void StopService()
		{
			_Timer.Enabled = false;
			foreach (IServerTimeNotification callback in _Clients.Values) {
				(callback as IDisposable).Dispose();
			}
			_Timer.Dispose();
		}

		public void Subscribe()
		{
			OperationContext ctx = OperationContext.Current;
			if (_Clients.ContainsKey(ctx.SessionId) == false) {
				IServerTimeNotification callback = ctx.GetCallbackChannel<IServerTimeNotification>();
				lock (_Clients) {
					_Clients.Add(ctx.SessionId, callback);
				}
				Console.WriteLine("Event subscription : SessionId {0}", ctx.SessionId);
			}
		}

		public void Unsubscribe()
		{
			OperationContext ctx = OperationContext.Current;
			if (_Clients.ContainsKey(ctx.SessionId) == true) {
				lock (_Clients) {
					_Clients.Remove(ctx.SessionId);
				}
				Console.WriteLine("Event unsubscription : SessionId {0}", ctx.SessionId);
			}
		}
	}

	class ServiceApp
	{
		static void Main(string[] args)
		{
			Console.WriteLine("Event Sample Service using Callback......");

			WSDualHttpBinding binding = new WSDualHttpBinding();
			binding.Security.Mode = WSDualHttpSecurityMode.None;

			ServiceMetadataBehavior behavior = new ServiceMetadataBehavior();
			behavior.HttpGetEnabled = true;

			ServiceHost host = new ServiceHost(
				typeof(ServerTimeService),
				new Uri("http://localhost/wcf/example/callback/servertime"));
			host.AddServiceEndpoint(typeof(IServerTime), binding, "");
			host.Description.Behaviors.Add(behavior);

			host.Open();
			ServerTimeService.StartService();
			Console.WriteLine("ServerTimeService started....");
			Console.ReadKey(true);
			ServerTimeService.StopService();
			host.Close();
		}
	}
}
